////////////////////////////////////////
//(c) 1999, 2000 EXAR Corporation 
//All Rights Reserved.
//
//Exar Corporation ("Exar") hereby grants the User of this Exar 
//product-specific XR17C158 Windows NT Device Driver (hereinafter 
//referred to as "Driver") a non-exclusive, nontransferable license 
//to use the Driver in accordance with the term set forth below.  
//Before using the Driver, the User should read the following use 
//restrictions.  If the User does not accept these terms, the Driver 
//should not be used or downloaded.
//The User is granted this limited license to use this Driver for 
//User's end product that includes the Exar XR17C158 Octal UART device, 
//and is not granted rights to sell, loan, rent, lease or license the 
//Driver, in whole or in part, or in modified form to anyone other 
//than User.  User may modify the Driver to suit its specific 
//applications but rights to derivative works and such modifications 
//shall belong to Exar.
//
//The Driver is provided on an "AS IS" basis and Exar makes absolutely 
//no warranty with respect to performance or the information contained 
//therein.  EXAR DISCLAIMS AND USER WAIVES ALL WARRANTIES, EXPRESS OR 
//IMPLIED, INCLUDING WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A 
//PARTICULAR PURPOSE.  The entire risk as to Driver quality and 
//performance is with the User.  ACCORDINGLY, IN NO EVENT SHALL EXAR 
//BE LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT, INCLUDING 
//ANY LOST PROFITS OR OTHER INCIDENTAL, CONSEQUENTIAL, EXEMPLARY, OR 
//PUNITIVE DAMAGES ARISING OUT OF THE USE OR APPLICATION OF THE DRIVER.  
//Further, Exar reserves the right to make changes tot eh Driver without 
//notice to improve reliability, function or design.  Exar does not 
//convey any license under patent rights or any other intellectual 
//property rights, including those of third parties.
//
//Exar is not obligated to provide maintenance or support for the Driver. 
//
//Author:	Kevin Guo
//
//////////////////////////////////////////////////

#include "stdafx.h"

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <conio.h>

#include "hardware.h"
#include "ioaccess.h"

//#define KGTIME0	0x1999
#define KGTIME0	0xEFFF

CHardWare::CHardWare(DWORD dwAddr, HANDLE hDriverHandle) 
{
	dwMemBase = dwAddr;
	dwOffset = 0x8e;
	byOldState = 0;
	hDriver = hDriverHandle;
}

KGBYTE
CHardWare::Read()
{
#define WIN2K
#ifdef WIN2K
	DWORD  cbReturned;
	BYTE byValue;
	if (!DeviceIoControl(hDriver,
				 IOCTL_XRPORT_READ_REG2,
				 NULL,
				 0,
				 &byValue,
				 sizeof(BYTE),
				 &cbReturned,
				 0
				 ))
	{
		byValue = 0;
		MessageBox(NULL, "Error in Reading...!", "Error", MB_OK);
	}	
	return byValue;
#endif
	
	KGBYTE byVal = READ_REGISTER_UCHAR(dwMemBase + dwOffset);
	return byVal;
}

VOID
CHardWare::Write( KGBYTE byValue)
{
#define WIN2K
#ifdef WIN2K

	DWORD  cbReturned;
	if (!DeviceIoControl(hDriver,
				 (DWORD) IOCTL_XRPORT_WRITE_REG2,
				 &byValue,
				 sizeof(BYTE),
				 NULL,
				 0,
				 &cbReturned,
				 0
				 ))
	{		
		MessageBox(NULL, "Error in Writing...!", "Error", MB_OK);
	}
	return;
#endif

	WRITE_REGISTER_UCHAR(dwMemBase + dwOffset, byValue);
}

VOID
CHardWare::WriteClk( BOOL bValue )
{
	KGBYTE byVal;
	byVal = byOldState;
	if(bValue)
	{
		byVal |= 0x10;
		Write(byVal);
		byOldState |= 0x10;
	}
	else
	{
		byVal &= ~0x10;
		Write(byVal);
		byOldState &= ~0x10;
	}
}

VOID
CHardWare::WriteCs( BOOL bValue )
{
	KGBYTE byVal;
	byVal = byOldState;
	if(bValue)
	{
		byVal |= 0x20;
		Write(byVal);
		byOldState |= 0x20;
	}
	else
	{
		byVal &= ~0x20;
		Write(byVal);
		byOldState &= ~0x20;
	}
}

VOID
CHardWare::WriteDi( BOOL bValue )
{
	KGBYTE byVal;
	byVal = byOldState;
	if(bValue)
	{
		byVal |= 0x40;
		Write(byVal);
		byOldState |= 0x40;
	}
	else
	{
		byVal &= ~0x40;
		Write(byVal);
		byOldState &= ~0x40;
	}
}

KGBYTE
CHardWare::ReadDo()
{
	KGBYTE byVal;
	byVal = Read();
	byVal &= 0x80;
	return byVal;
}

DWORD
CHardWare::ReadE2prom( DWORD dwAddrInRom )
{
	DWORD dwTemp = 99;
	KGBYTE byTemp = 0;
	int  dwRead = 0;

	dwAddrInRom = ReOrder(dwAddrInRom, 0x06);

	WriteCs(HIGH);
	KgSleep(KGTIME0);
	WriteDi(HIGH);  //DI = 1
	KgSleep(KGTIME0);

	WriteClk(HIGH);
	KgSleep(KGTIME0);
	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteClk(HIGH);  //DI = 1
	KgSleep(KGTIME0);
	WriteClk(LOW);
	WriteDi(LOW);
	KgSleep(KGTIME0);
	
	WriteClk(HIGH);  //DI = 0
	KgSleep(KGTIME0);

	for(DWORD i=0; i<6; i++)
	{
		dwTemp = dwAddrInRom & 0x01;
		WriteClk(LOW);
		WriteDi((BOOL) dwTemp);
		dwAddrInRom = dwAddrInRom >> 1;

		KgSleep(KGTIME0);
		WriteClk(HIGH);
		KgSleep(KGTIME0);
	}

	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteClk(HIGH);
	KgSleep(KGTIME0);
	WriteClk(LOW);
	KgSleep(KGTIME0);


	//Read D0 out
	for(DWORD j=0; j<16; j++)
	{
		WriteClk(HIGH);
		byTemp = ReadDo();
		if(byTemp)
		{
			dwRead |= 0x01;
			dwRead = dwRead << 1;
		}
		else
		{
			dwRead &= 0xfffe;
			dwRead = dwRead << 1;
		}
		KgSleep(KGTIME0);
		WriteClk(LOW);
		KgSleep(KGTIME0);
	}
	WriteCs(LOW);
	KgSleep(KGTIME0);
	dwRead = dwRead >> 1;

//	dwRead = 0;
	return (DWORD) dwRead;
}

VOID
CHardWare::WriteE2prom( DWORD dwAddrInRom, DWORD dwData )
{
	DWORD dwTemp = 99;
	KGBYTE byTemp = 0;

	dwAddrInRom = ReOrder(dwAddrInRom, 0x06);

	WriteCs(HIGH);
	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteDi(HIGH);  //DI = 1
	KgSleep(KGTIME0);

	WriteClk(HIGH);
	KgSleep(KGTIME0);
	WriteClk(LOW);

	WriteDi(LOW);
	KgSleep(KGTIME0);
	WriteClk(HIGH);  //DI = 0
	KgSleep(KGTIME0);
	WriteClk(LOW);
	WriteDi(HIGH);
	KgSleep(KGTIME0);
	
	WriteClk(HIGH);  //DI = 1
	KgSleep(KGTIME0);

	for(DWORD i=0; i<6; i++)
	{
		dwTemp = dwAddrInRom & 0x01;
		WriteClk(LOW);
		WriteDi((BOOL) dwTemp);
		dwAddrInRom = dwAddrInRom >> 1;

		KgSleep(KGTIME0);
		WriteClk(HIGH);
		KgSleep(KGTIME0);
	}

	dwData = ReOrder(dwData, 16);
	for( i=0; i<16; i++)
	{
		dwTemp = dwData & 0x01;
		WriteClk(LOW);
		WriteDi((BOOL) dwTemp);
		dwData = dwData >> 1;

		KgSleep(KGTIME0);
		WriteClk(HIGH);
		KgSleep(KGTIME0);
	}
	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteCs(LOW);
	KgSleep(KGTIME0);
}

VOID
CHardWare::WriteEnable( BOOL bTrue )
{
	WriteCs(HIGH);
	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteDi(HIGH);  //DI = 1
	KgSleep(KGTIME0);

	WriteClk(HIGH);
	KgSleep(KGTIME0);
	WriteClk(LOW);

	WriteDi(LOW);
	KgSleep(KGTIME0);
	WriteClk(HIGH);  //DI = 0
	KgSleep(KGTIME0);
	WriteClk(LOW);
	WriteDi(LOW);
	KgSleep(KGTIME0);
	
	WriteClk(HIGH);  //DI = 0
	KgSleep(KGTIME0);

	if(bTrue)
	{
		WriteClk(LOW);
		WriteDi(HIGH);
		KgSleep(KGTIME0);
		WriteClk(HIGH);  //DI = 1
		KgSleep(KGTIME0);
		
		WriteClk(LOW);
		WriteDi(HIGH);
		KgSleep(KGTIME0);
		WriteClk(HIGH);  //DI = 1
		KgSleep(KGTIME0);
	}
	else
	{
		WriteClk(LOW);
		WriteDi(LOW);
		KgSleep(KGTIME0);
		WriteClk(HIGH);  //DI = 1
		KgSleep(KGTIME0);
		
		WriteClk(LOW);
		WriteDi(LOW);
		KgSleep(KGTIME0);
		WriteClk(HIGH);  //DI = 1
		KgSleep(KGTIME0);
	}

	for(DWORD i=0; i<5; i++)
	{
		WriteClk(LOW);
		KgSleep(KGTIME0);
		WriteClk(HIGH); 
		KgSleep(KGTIME0);
	}

	WriteClk(LOW);
	KgSleep(KGTIME0);
	WriteCs(LOW);
	KgSleep(KGTIME0);
}

DWORD
CHardWare::ReOrder(DWORD dwData, KGBYTE byPos)
{
	DWORD dwTemp = 0;
	DWORD dwTemp2 = 0;
	DWORD byBit = 0;
	
	if(byPos == 6)
		byBit = 0x20;
	else if(byPos == 16)
		byBit = 0x8000;
	else
		byBit = 0;
	
	for(KGBYTE i=0; i<byPos; i++)
	{
		dwTemp = dwData & byBit;
		dwData = dwData << 1;
		dwTemp = dwTemp >> (byPos-1-i);
		dwTemp2 |= dwTemp;
	}
	dwData = dwTemp2;
  	return dwData;
}

VOID
CHardWare::KgSleep(DWORD dwCount)
{	
	int temp;
	for(DWORD i=0; i<dwCount; i++)
	{
		temp = i;
	}
}